@@ -1,24 +1,14 @@
+// SPDX-License-Identifier: GPL-2.0
 /*
  * Copyright (c) 2003-2005 Silicon Graphics, Inc.
  * All Rights Reserved.
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation.
- *
- * This program is distributed in the hope that it would be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write the Free Software Foundation,
- * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <xfs/xfs.h>
-#include <xfs/command.h>
-#include <xfs/input.h>
+#define _BSD_SOURCE
+#define _DEFAULT_SOURCE
+#include <sys/uio.h>
+#include "command.h"
+#include "input.h"
 #include <ctype.h>
 #include "init.h"
 #include "io.h"
@@ -46,6 +36,10 @@ pread_help(void)
 " -R   -- read at random offsets in the range of bytes\n"
 " -Z N -- zeed the random number generator (used when reading randomly)\n"
 "         (heh, zorry, the -s/-S arguments were already in use in pwrite)\n"
+#ifdef HAVE_PREADV
+" -V N -- use vectored IO with N iovecs of blocksize each (preadv)\n"
+#endif
+"\n"
 " When in \"random\" mode, the number of read operations will equal the\n"
 " number required to do a complete forward/backward scan of the range.\n"
 " Note that the offset within the range is chosen at random each time\n"
@@ -53,9 +47,44 @@ pread_help(void)
 "\n"));
 }
 
-void	*buffer;
+void	*io_buffer;
 size_t	highwater;
-size_t	buffersize;
+size_t	io_buffersize;
+int	vectors;
+struct iovec *iov;
+
+static int
+alloc_iovec(
+	size_t		bsize,
+	int		uflag,
+	unsigned int	seed)
+{
+	int		i;
+
+	iov = calloc(vectors, sizeof(struct iovec));
+	if (!iov)
+		return -1;
+
+	io_buffersize = 0;
+	for (i = 0; i < vectors; i++) {
+		iov[i].iov_base = memalign(pagesize, bsize);
+		if (!iov[i].iov_base) {
+			perror("memalign");
+			goto unwind;
+		}
+		iov[i].iov_len = bsize;
+		if (!uflag)
+			memset(iov[i].iov_base, seed, bsize);
+	}
+	io_buffersize = bsize * vectors;
+	return 0;
+unwind:
+	for( ; i >= 0; i--)
+		free(iov[i].iov_base);
+	free(iov);
+	iov = NULL;
+	return -1;
+}
 
 int
 alloc_buffer(
@@ -63,32 +92,36 @@ alloc_buffer(
 	int		uflag,
 	unsigned int	seed)
 {
+	if (vectors)
+		return alloc_iovec(bsize, uflag, seed);
+
 	if (bsize > highwater) {
-		if (buffer)
-			free(buffer);
-		buffer = memalign(pagesize, bsize);
-		if (!buffer) {
+		if (io_buffer)
+			free(io_buffer);
+		io_buffer = memalign(pagesize, bsize);
+		if (!io_buffer) {
 			perror("memalign");
-			highwater = buffersize = 0;
+			highwater = io_buffersize = 0;
 			return -1;
 		}
 		highwater = bsize;
 	}
-	buffersize = bsize;
+	io_buffersize = bsize;
 	if (!uflag)
-		memset(buffer, seed, buffersize);
+		memset(io_buffer, seed, io_buffersize);
 	return 0;
 }
 
 void
-dump_buffer(
+__dump_buffer(
+	void		*buf,
 	off64_t		offset,
 	ssize_t		len)
 {
 	int		i, j;
 	char		*p;
 
-	for (i = 0, p = (char *)buffer; i < len; i += 16) {
+	for (i = 0, p = (char *)buf; i < len; i += 16) {
 		char	*s = p;
 
 		printf("%08llx:  ", (unsigned long long)offset + i);
@@ -105,34 +138,74 @@ dump_buffer(
 	}
 }
 
-int
-align_direct(
-	off64_t		*offset,
-	long long	*count)
+void
+dump_buffer(
+	off64_t		offset,
+	ssize_t		len)
 {
-	struct dioattr	dio;
-	unsigned int	align;
+	int		i, l;
 
-	if (!(file->flags & IO_DIRECT))
-		return 0;
-
-	if (file->flags & IO_FOREIGN) {
-		dio.d_miniosz = BBSIZE;	/* punt */
-	} else if ((xfsctl(file->name, file->fd, XFS_IOC_DIOINFO, &dio)) < 0) {
-		perror("XFS_IOC_DIOINFO");
-		return 1;
+	if (!vectors) {
+		__dump_buffer(io_buffer, offset, len);
+		return;
 	}
 
-	align = (*offset % dio.d_miniosz);
-	if (align) {
-		*offset -= align;
-		*count += align;
+	for (i = 0; len > 0 && i < vectors; i++) {
+		l = min(len, iov[i].iov_len);
+
+		__dump_buffer(iov[i].iov_base, offset, l);
+		len -= l;
+		offset += l;
 	}
-	align = (*count % dio.d_miniosz);
-	if (align) {
-		*count += (dio.d_miniosz - align);
+}
+
+#ifdef HAVE_PREADV
+static ssize_t
+do_preadv(
+	int		fd,
+	off64_t		offset,
+	long long	count)
+{
+	int		vecs = 0;
+	ssize_t		oldlen = 0;
+	ssize_t		bytes = 0;
+
+	/* trim the iovec if necessary */
+	if (count < io_buffersize) {
+		size_t	len = 0;
+		while (len + iov[vecs].iov_len < count) {
+			len += iov[vecs].iov_len;
+			vecs++;
+		}
+		oldlen = iov[vecs].iov_len;
+		iov[vecs].iov_len = count - len;
+		vecs++;
+	} else {
+		vecs = vectors;
 	}
-	return 0;
+	bytes = preadv(fd, iov, vectors, offset);
+
+	/* restore trimmed iov */
+	if (oldlen)
+		iov[vecs - 1].iov_len = oldlen;
+
+	return bytes;
+}
+#else
+#define do_preadv(fd, offset, count) (0)
+#endif
+
+static ssize_t
+do_pread(
+	int		fd,
+	off64_t		offset,
+	long long	count,
+	size_t		buffer_size)
+{
+	if (!vectors)
+		return pread(fd, io_buffer, min(count, buffer_size), offset);
+
+	return do_preadv(fd, offset, count);
 }
 
 static int
@@ -149,29 +222,33 @@ read_random(
 	int		ops = 0;
 
 	srandom(seed);
-	end = lseek64(fd, 0, SEEK_END);
+	end = lseek(fd, 0, SEEK_END);
 	offset = (eof || offset > end) ? end : offset;
-	if ((bytes = (offset % buffersize)))
+	if ((bytes = (offset % io_buffersize)))
 		offset -= bytes;
 	offset = max(0, offset);
-	if ((bytes = (count % buffersize)))
+	if ((bytes = (count % io_buffersize)))
 		count += bytes;
-	count = max(buffersize, count);
-	range = count - buffersize;
+	count = max(io_buffersize, count);
+	range = count - io_buffersize;
 
 	*total = 0;
 	while (count > 0) {
-		off = ((random() % range) / buffersize) * buffersize;
-		bytes = pread64(fd, buffer, buffersize, off);
+		if (range)
+			off = ((offset + (random() % range)) / io_buffersize) *
+				io_buffersize;
+		else
+			off = offset;
+		bytes = do_pread(fd, off, io_buffersize, io_buffersize);
 		if (bytes == 0)
 			break;
 		if (bytes < 0) {
-			perror("pread64");
+			perror("pread");
 			return -1;
 		}
 		ops++;
 		*total += bytes;
-		if (bytes < buffersize)
+		if (bytes < io_buffersize)
 			break;
 		count -= bytes;
 	}
@@ -191,8 +268,8 @@ read_backward(
 	long long	cnt = *count;
 	int		ops = 0;
 
-	end = lseek64(fd, 0, SEEK_END);
-	off = eof ? end : min(end, lseek64(fd, off, SEEK_SET));
+	end = lseek(fd, 0, SEEK_END);
+	off = eof ? end : min(end, lseek(fd, off, SEEK_SET));
 	if ((end = off - cnt) < 0) {
 		cnt += end;	/* subtraction, end is negative */
 		end = 0;
@@ -202,14 +279,13 @@ read_backward(
 	*offset = off;
 
 	/* Do initial unaligned read if needed */
-	if ((bytes_requested = (off % buffersize))) {
-		bytes_requested = min(cnt, bytes_requested);
+	if ((bytes_requested = (off % io_buffersize))) {
 		off -= bytes_requested;
-		bytes = pread(fd, buffer, bytes_requested, off);
+		bytes = do_pread(fd, off, bytes_requested, io_buffersize);
 		if (bytes == 0)
 			return ops;
 		if (bytes < 0) {
-			perror("pread64");
+			perror("pread");
 			return -1;
 		}
 		ops++;
@@ -221,13 +297,13 @@ read_backward(
 
 	/* Iterate backward through the rest of the range */
 	while (cnt > end) {
-		bytes_requested = min(cnt, buffersize);
+		bytes_requested = min(cnt, io_buffersize);
 		off -= bytes_requested;
-		bytes = pread64(fd, buffer, bytes_requested, off);
+		bytes = do_pread(fd, off, cnt, io_buffersize);
 		if (bytes == 0)
 			break;
 		if (bytes < 0) {
-			perror("pread64");
+			perror("pread");
 			return -1;
 		}
 		ops++;
@@ -249,25 +325,23 @@ read_forward(
 	int		onlyone,
 	int		eof)
 {
-	size_t		bytes_requested;
 	ssize_t		bytes;
 	int		ops = 0;
 
 	*total = 0;
 	while (count > 0 || eof) {
-		bytes_requested = min(count, buffersize);
-		bytes = pread64(fd, buffer, bytes_requested, offset);
+		bytes = do_pread(fd, offset, count, io_buffersize);
 		if (bytes == 0)
 			break;
 		if (bytes < 0) {
-			perror("pread64");
+			perror("pread");
 			return -1;
 		}
 		ops++;
 		if (verbose)
 			dump_buffer(offset, bytes);
 		*total += bytes;
-		if (onlyone || bytes < bytes_requested)
+		if (onlyone || bytes < min(count, io_buffersize))
 			break;
 		offset += bytes;
 		count -= bytes;
@@ -298,7 +372,6 @@ pread_f(
 	long long	count, total, tmp;
 	size_t		fsblocksize, fssectsize;
 	struct timeval	t1, t2;
-	char		s1[64], s2[64], ts[64];
 	char		*sp;
 	int		Cflag, qflag, uflag, vflag;
 	int		eof = 0, direction = IO_FORWARD;
@@ -308,12 +381,13 @@ pread_f(
 	init_cvtnum(&fsblocksize, &fssectsize);
 	bsize = fsblocksize;
 
-	while ((c = getopt(argc, argv, "b:BCFRquvZ:")) != EOF) {
+	while ((c = getopt(argc, argv, "b:BCFRquvV:Z:")) != EOF) {
 		switch (c) {
 		case 'b':
 			tmp = cvtnum(fsblocksize, fssectsize, optarg);
 			if (tmp < 0) {
 				printf(_("non-numeric bsize -- %s\n"), optarg);
+				exitcode = 1;
 				return 0;
 			}
 			bsize = tmp;
@@ -339,25 +413,41 @@ pread_f(
 		case 'v':
 			vflag = 1;
 			break;
+#ifdef HAVE_PREADV
+		case 'V':
+			vectors = strtoul(optarg, &sp, 0);
+			if (!sp || sp == optarg) {
+				printf(_("non-numeric vector count == %s\n"),
+					optarg);
+				exitcode = 1;
+				return 0;
+			}
+			break;
+#endif
 		case 'Z':
 			zeed = strtoul(optarg, &sp, 0);
 			if (!sp || sp == optarg) {
 				printf(_("non-numeric seed -- %s\n"), optarg);
+				exitcode = 1;
 				return 0;
 			}
 			break;
 		default:
+			exitcode = 1;
 			return command_usage(&pread_cmd);
 		}
 	}
-	if (optind != argc - 2)
+	if (optind != argc - 2) {
+		exitcode = 1;
 		return command_usage(&pread_cmd);
+	}
 
 	offset = cvtnum(fsblocksize, fssectsize, argv[optind]);
 	if (offset < 0 && (direction & (IO_RANDOM|IO_BACKWARD))) {
 		eof = -1;	/* read from EOF */
 	} else if (offset < 0) {
 		printf(_("non-numeric length argument -- %s\n"), argv[optind]);
+		exitcode = 1;
 		return 0;
 	}
 	optind++;
@@ -366,13 +456,14 @@ pread_f(
 		eof = -1;	/* read to EOF */
 	} else if (count < 0) {
 		printf(_("non-numeric length argument -- %s\n"), argv[optind]);
+		exitcode = 1;
 		return 0;
 	}
 
-	if (alloc_buffer(bsize, uflag, 0xabababab) < 0)
-		return 0;
-	if (align_direct(&offset, &count) < 0)
+	if (alloc_buffer(bsize, uflag, 0xabababab) < 0) {
+		exitcode = 1;
 		return 0;
+	}
 
 	gettimeofday(&t1, NULL);
 	switch (direction) {
@@ -392,40 +483,30 @@ pread_f(
 	default:
 		ASSERT(0);
 	}
-	if (c < 0)
+	if (c < 0) {
+		exitcode = 1;
 		return 0;
+	}
+
 	if (qflag)
 		return 0;
 	gettimeofday(&t2, NULL);
 	t2 = tsub(t2, t1);
 
-	/* Finally, report back -- -C gives a parsable format */
-	timestr(&t2, ts, sizeof(ts), Cflag ? VERBOSE_FIXED_TIME : 0);
-	if (!Cflag) {
-		cvtstr((double)total, s1, sizeof(s1));
-		cvtstr(tdiv((double)total, t2), s2, sizeof(s2));
-		printf(_("read %lld/%lld bytes at offset %lld\n"),
-			total, count, (long long)offset);
-		printf(_("%s, %d ops; %s (%s/sec and %.4f ops/sec)\n"),
-			s1, c, ts, s2, tdiv((double)c, t2));
-	} else {/* bytes,ops,time,bytes/sec,ops/sec */
-		printf("%lld,%d,%s,%.3f,%.3f\n",
-			total, c, ts,
-			tdiv((double)total, t2), tdiv((double)c, t2));
-	}
+	report_io_times("read", &t2, (long long)offset, count, total, c, Cflag);
 	return 0;
 }
 
 void
 pread_init(void)
 {
-	pread_cmd.name = _("pread");
-	pread_cmd.altname = _("r");
+	pread_cmd.name = "pread";
+	pread_cmd.altname = "r";
 	pread_cmd.cfunc = pread_f;
 	pread_cmd.argmin = 2;
 	pread_cmd.argmax = -1;
 	pread_cmd.flags = CMD_NOMAP_OK | CMD_FOREIGN_OK;
-	pread_cmd.args = _("[-b bs] [-v] off len");
+	pread_cmd.args = _("[-b bs] [-v] [-i N] [-FBR [-Z N]] off len");
 	pread_cmd.oneline = _("reads a number of bytes at a specified offset");
 	pread_cmd.help = pread_help;
 
